#!/usr/bin/env python
# encoding: utf8

""" 
Components setup and definitions.
"""

from __future__ import (absolute_import, print_function, division, unicode_literals)



#
# Some basic constants for convenience
#

#: Colors for figures
PATH_COLORS = {
    'piccadillyblue': (0 / 255, 25 / 255, 168 / 255),
    'centralred': (220 / 255, 36 / 255, 31 / 255),
    'districtgreen': (0 / 255, 114 / 255, 41 / 255),
    'circleyellow': (255 / 255, 206 / 255, 0 / 255),
    'jubileegray': (134 / 255, 143 / 255, 152 / 255),
    'metropolitangmagenta': (117 / 255, 16 / 255, 86 / 255),
    'safetyblue': (0 / 255, 103 / 255, 193 / 255),
    'riverservicesblue': (0 / 255, 160 / 255, 226 / 255),
    'congestionred': (210 / 255, 16 / 255, 52 / 255),
    'hammersmithpink': (215 / 255, 153 / 255, 175 / 255),
    'victoriablue': (0 / 255, 160 / 255, 226 / 255),
    'waterloogreen': (118 / 255, 208 / 255, 189 / 255),
    'bakerloobrown': (137 / 255, 78 / 255, 36 / 255),
    'overgroundorange': (232 / 255, 106 / 255, 16 / 255),
    'dlrteal': (0 / 255, 175 / 255, 173 / 255)}



#
# Define game components
#

# Lists of available game components
_LOCATIONS = dict()
_ROCKETS = dict()
_PROBES = dict()


# Convenience functions for retrieving components by name; hides the implementation the lists above
#

def get_rocket(name):
    """Get a rocket component by name."""
    return _ROCKETS[name]


def get_probe(name):
    """Get a probe component by name."""
    return _PROBES[name]


def get_component(name):
    """Get a rocket or probe component by name."""
    if name in PROBES_BASIC:
        return get_probe(name)
    elif name in ROCKETS_BASIC:
        return get_rocket(name)
    else:
        return None


def get_location(name):
    """Get a location by name."""
    return _LOCATIONS[name]


# Internal functions for instantiating and saving and connecting game components
#

def _add_rocket(name, mass=0, cost=0, thrust=0):
    assert name not in _ROCKETS.keys(), 'ERROR: You cannot add a rocket with the same name as an existing one.'
    _ROCKETS[name] = Rocket(name, mass, cost, thrust)


def _add_probe(name, mass=0, cost=0, crew=0):
    assert name not in _PROBES.keys(), 'ERROR: You cannot add a probe with the same name as an existing one.'
    _PROBES[name] = Probe(name, mass, cost, crew)


def _add_location(name):
    assert name not in _LOCATIONS.keys(), 'ERROR: You cannot add a location with the same name as an existing one.'
    _LOCATIONS[name] = Location(name)


def _connect_locations(start, end, difficulty):
    _LOCATIONS[start].connections += [Connection(get_location(end), difficulty)]
    # REV - Alternate using dict (see get_connection_to)
    # _LOCATIONS[start].connections[end] = Connection(get_location(end), difficulty)


# Definitions of components
#

class Component(object):
    """The Component."""
    def __init__(self, name, mass=0, cost=0):
        #: The name
        self._name = name
        self._mass = mass
        self._cost = cost

    @property
    def name(self):
        """The name of a component.

        Returns:
            unicode: A string uniquely specifying a `Component`.

        """
        return self._name

    @property
    def mass(self):
        """The mass of a component.

        Returns:
            int: The mass of the `Component`.

        """
        return self._mass

    @property
    def cost(self):
        """The cost of a component.

        Returns:
            int: The cost of the `Component`.

        """
        return self._cost

class Rocket(Component):

    def __init__(self, name, mass=0, cost=0, thrust=0):
        super(Rocket, self).__init__(name, mass, cost)
        self.thrust = thrust


class Probe(Component):

    def __init__(self, name, mass=0, cost=0, crew=0):
        super(Probe, self).__init__(name, mass, cost)
        self.crew = crew


class Location(object):

    def __init__(self, name):
        self.name = name
        self.connections = []

    def get_connection_to(self, end_location):
        end = next((con for con in self.connections if con.end == end_location), None)
        assert end is not None, 'ERROR: {} is not reachable from {}!'.format(end_location.name, self.name)
        return end
        # REV - Alternate using dict (see _connect_locations)
        # assert end_location.name in self.connections.keys(), 'ERROR: {} is not reachable from {}!'.format(end_location.name, self.name)
        # return self.connections[end_location.name]


class Connection(object):
    def __init__(self, end, difficulty):
        # TBD - Restore?
        # self.start = start
        self.end = end
        self.difficulty = difficulty



#
# Build game components
#

for loc_name in ['Earth', 'Suborbital Flight', 'Earth Orbit', 'Lunar Orbit', 'Lunar Fly-By', 'Moon',
                 'Inner Planets Transfer',
                 'Ceres',
                 'Venus Orbit', 'Venus Fly-By', 'Venus',
                 'Mars Orbit', 'Mars Fly-By', 'Mars', 'Phobos']:
    _add_location(loc_name)

_connect_locations('Earth', 'Suborbital Flight', difficulty=3)
_connect_locations('Earth', 'Earth Orbit', difficulty=8)
_connect_locations('Suborbital Flight', 'Earth', difficulty=-1)
_connect_locations('Suborbital Flight', 'Earth Orbit', difficulty=5)
_connect_locations('Earth Orbit', 'Earth', difficulty=0)
_connect_locations('Earth Orbit', 'Lunar Fly-By', difficulty=1)
_connect_locations('Earth Orbit', 'Lunar Orbit', difficulty=3)
_connect_locations('Earth Orbit', 'Inner Planets Transfer', difficulty=3)
_connect_locations('Earth Orbit', 'Mars Fly-By', difficulty=3)
_connect_locations('Earth Orbit', 'Mars Orbit', difficulty=5)
#_connect_locations('Earth Orbit', 'Outer Planets', difficulty=6)
#
_connect_locations('Lunar Fly-By', 'Earth Orbit', difficulty=1)
_connect_locations('Lunar Fly-By', 'Lunar Orbit', difficulty=2)
_connect_locations('Lunar Fly-By', 'Moon', difficulty=4)
_connect_locations('Lunar Orbit', 'Earth Orbit', difficulty=3)
_connect_locations('Lunar Orbit', 'Moon', difficulty=2)
_connect_locations('Moon', 'Lunar Orbit', difficulty=2)

_connect_locations('Inner Planets Transfer', 'Earth Orbit', difficulty=3)
_connect_locations('Inner Planets Transfer', 'Venus Fly-By', difficulty=2)
_connect_locations('Inner Planets Transfer', 'Venus Orbit', difficulty=3)
_connect_locations('Inner Planets Transfer', 'Ceres', difficulty=5)
_connect_locations('Inner Planets Transfer', 'Mars Orbit', difficulty=4)

_connect_locations('Ceres', 'Inner Planets Transfer', difficulty=5)

_connect_locations('Venus Orbit', 'Inner Planets Transfer', difficulty=3)
#_connect_locations('Venus Orbit', 'Outer Planets', difficulty=9)
_connect_locations('Venus Fly-By', 'Venus Orbit', difficulty=1)
_connect_locations('Venus Fly-By', 'Venus', difficulty=1)

_connect_locations('Mars Orbit', 'Earth Orbit', difficulty=5)
_connect_locations('Mars Orbit', 'Inner Planets Transfer', difficulty=4)
#_connect_locations('Mars Orbit', 'Outer Planets', difficulty=5)
_connect_locations('Mars Orbit', 'Phobos', difficulty=1)
_connect_locations('Mars Orbit', 'Mars', difficulty=0)
_connect_locations('Mars Fly-By', 'Mars Orbit', difficulty=3)
_connect_locations('Mars Fly-By', 'Mars', difficulty=3)
_connect_locations('Mars', 'Mars Orbit', difficulty=3)
_connect_locations('Phobos', 'Mars Orbit', difficulty=1)

_add_rocket('Juno', 1, 1, 4),
_add_rocket('Atlas', 4, 5, 27),
_add_rocket('Soyuz', 9, 8, 80),
_add_rocket('Saturn', 20, 15, 200)

ROCKETS_BASIC = _ROCKETS.keys()

_add_probe('Probe', 1, 2, 0)
_add_probe('Eagle', 1, 4, 2)
_add_probe('Vostok', 2, 2, 1)
_add_probe('Apollo', 3, 4, 3)
_add_probe('Aldrin', 3, 4, 8)
# TBD - Placeholder to simulate samples <<<
_add_probe('Sample', 1, 0, 0)

PROBES_BASIC = _PROBES.keys()

PAYLOADS = PROBES_BASIC
COMPONENTS = PAYLOADS + ROCKETS_BASIC

ALL_ROCKETS = ROCKETS_BASIC




