#!/usr/bin/env python
# encoding: utf8

""" 
Tools for mission planning.
"""

from __future__ import (absolute_import, print_function, division, unicode_literals)

import matplotlib
# matplotlib.use('agg') -- http://matplotlib.org/faq/usage_faq.html#what-is-a-backend
import matplotlib.pyplot
import matplotlib.patheffects
import itertools

from .components import *


class Stage(object):
    """A stage of the mission, consisting of a location; and any burns, additions, and subtractions there.

    A valid Stage consists of a location and a two lists of tuples, each of which specifies a Component and
    a number. These indicate components added or removed at that location as part of a mission (but not
    burned there)."""

    def __init__(self, location_name, additions=[], subtractions=[]):
        self.location = get_location(location_name)
        self.additions = [(get_component(s_name), n) for (s_name, n) in additions]
        self.subtractions = [(get_component(a_name), n) for (a_name, n) in subtractions]
        self.burns = []  # Burns to get to next location in a Mission

    def cost(self):
        return sum(r.cost * n for (r, n) in self.burns)

    def __str__(self):
        return 'location: {}; burns: {}'.format(self.location, self.burns)


class Mission(object):
    # Path is entered in order, but reversed throughout for convenience of computation and display
    def __init__(self, path, userockets=ROCKETS_BASIC, excluderockets=[]):
        self.stages = [Stage(*spec) if isinstance(spec, tuple) else Stage(spec) for spec in reversed(path)]
        self.userockets = userockets
        self.excluderockets = excluderockets
        self.rockets = None
        self.compute_burns(userockets=userockets, excluderockets=excluderockets)

    def cost(self):
        return {'payload': 0, 'rockets': sum(self.rockets[r] * r.cost for r in self.rockets.keys()), 'total': 0}

    def compute_burns(self, userockets=ROCKETS_BASIC, excluderockets=[]):
        self.rockets = {}
        for s in self.stages:
            s.burns = []
        self._compute_burn([get_rocket(r) for r in ALL_ROCKETS if r in userockets and r not in excluderockets])

    def _compute_burn(self, rockets, prev_step=1):
        if prev_step >= len(self.stages):
            return
        else:
            payload_mass = sum(sum(c.mass * n for c, n in s.additions) for s in self.stages[prev_step:len(self.stages)])
            payload_mass -= sum(sum(c.mass * n for c, n in s.subtractions)
                                for s in self.stages[prev_step:len(self.stages)])
            rocket_mass = sum(sum(r.mass * n for r, n in s.burns) for s in self.stages[0:prev_step])
            difficulty = self.stages[prev_step].location.get_connection_to(
                self.stages[prev_step - 1].location).difficulty

            # TBD - Do something about hardcoding of range; replace with a loop?? <<<
            combos = (zip(rockets, ns) for ns in list(itertools.product(range(11), repeat=len(rockets))))
            combos = [c for c in combos if
                      sum(r.thrust * n for (r, n) in c) >= difficulty * (payload_mass + rocket_mass +
                                                                         sum(r.mass * n for (r, n) in c))]
            # REV - Improve optimization algorithm to handle cases low cost at one stage increases total cost later <<<
            if combos:
                # TBD - Filter out rockets with 0 count?
                self.stages[prev_step].burns = min(combos, key=lambda c: sum(r.cost * n for (r, n) in c))
                for r, n in self.stages[prev_step].burns:
                    if r in self.rockets.keys():
                        self.rockets[r] += n
                    else:
                        self.rockets[r] = n
            else:
                return

            self._compute_burn(rockets, prev_step + 1)

    def __str__(self):
        """Generate a basic string representation suitable for printing."""
        path_str = self.stages[0].location.name
        for s in self.stages[1:]:
            path_str += '\n^ '
            if s.burns != []:
                for (r, n) in s.burns:
                    if n > 0:
                        path_str += '{}x{} '.format(n, r.name)
                if s.cost() > 0:
                    path_str += '({})'.format(s.cost())
            else:
                path_str += 'XXXXX'
            path_str += '\n' + s.location.name
            if s.additions != []:
                path_str += ' +'
                for a, n in s.additions:
                    path_str += ' {}x{}'.format(n, a.name)
            if s.subtractions != []:
                path_str += ' -'
                for a, n in s.subtractions:
                    path_str += ' {}x{}'.format(n, a.name)
        return str(path_str)

    def plot(self, theme='Tube', path_color=None):
        """Generate simple graphical representation of the mission."""

        # TBD - Expose themes so they can be customized?
        THEMES = {
            'Tube': {'font': 'Arial', 'pathcolor': PATH_COLORS['piccadillyblue'], 'pathwidth': 4,
                     'markend': '_', 'marktransfers': 'o', 'markother': matplotlib.markers.TICKRIGHT,
                     'markadd': '►', 'marksubtract': '◄',
                     'arrowstyle': '-', 'failstyle': (0, (1, 3)), 'arrowshrink': 0.0},
            'xkcd': {'font': 'Courier New', 'pathcolor': 'k', 'pathwidth': 1,
                      'markend': '_', 'marktransfers': '_', 'markother': '_',
                      'markadd': '++', 'marksubtract': '--',
                      'arrowstyle': '->', 'failstyle': (0, (1, 4)), 'arrowshrink': 4.0},
            'Retro': {'font': 'Courier New', 'pathcolor': 'k', 'pathwidth': 1,
                      'markend': '_', 'marktransfers': '_', 'markother': '_',
                      'markadd': '++', 'marksubtract': '--',
                      'arrowstyle': '->', 'failstyle': (0, (1, 4)), 'arrowshrink': 4.0}}
        figure_font = THEMES[theme]['font']
        path_color = THEMES[theme]['pathcolor'] if path_color is None else path_color
        path_width = THEMES[theme]['pathwidth']
        out_arrow = dict(arrowstyle=THEMES[theme]['arrowstyle'], capstyle='projecting', joinstyle='bevel',
                         shrinkA=THEMES[theme]['arrowshrink'], shrinkB=THEMES[theme]['arrowshrink'],
                         color=path_color, linewidth=path_width)

        if theme == 'xkcd':
            # BUG - This causes subsequent figures to have some xkcd features; clearly not how this is done <<<
            matplotlib.pyplot.xkcd()

        fig, ax = matplotlib.pyplot.subplots()
        matplotlib.pyplot.axis('off')

        # REV - An alternate way to do the above
        # fig = matplotlib.pyplot.figure()
        # ax = fig.add_axes([0, 0, 1, 1], frameon=False)
        # ax.get_xaxis().set_visible(False)
        # ax.get_yaxis().set_visible(False)

        n_stages = len(self.stages)

        ax.set_ylim([-0.2, n_stages - 1 + 0.2])
        xy = [(0, 0)]

        for i, s in zip(range(n_stages), reversed(self.stages)):
            if s.additions == [] and s.subtractions == []:
                ax.plot(0, i,
                        marker=THEMES[theme]['markend'] if i in [0, n_stages - 1] else THEMES[theme][
                            'markother'],
                        color=path_color, mew=path_width,
                        ms=12 if i in [0, n_stages - 1] else 6, clip_on=False)
            else:
                ax.plot(0, i, marker=THEMES[theme]['marktransfers'],
                        color='w', mec='k', ms=8, mew=min(2, path_width), clip_on=False)
            ax.annotate(s.location.name, xy=(0, i), xytext=(10, 0), textcoords='offset points', ha='left', va='center',
                        fontsize=9, fontweight=600,
                        family=figure_font)
            transfers_label_str = ''
            if s.additions != []:
                for a, n in s.additions:
                    transfers_label_str += ' {} {}'.format(n, a.name)
                transfers_label_str += ' ' + THEMES[theme]['markadd']
            if s.subtractions != []:
                transfers_label_str += '\n' if s.additions != [] else ''
                for a, n in s.subtractions:
                    transfers_label_str += ' {} {}'.format(n, a.name)
                transfers_label_str += ' ' + THEMES[theme]['marksubtract']
            ax.annotate(transfers_label_str, xy=(0, i), xytext=(-10, 0), textcoords='offset points', fontsize=7,
                        family=THEMES[theme]['font'],
                        ha='right', va='center')

            if i < n_stages - 1:
                edge_label_str = ''
                cost_label_str = ''
                if s.burns != []:
                    for (r, n) in s.burns:
                        if n > 0:
                            edge_label_str += '{} {}\n'.format(n, r.name)
                    edge_label_str = edge_label_str[:-1]
                    cost_label_str += '[{}]'.format(s.cost())

                ax.annotate(edge_label_str, xy=(0, i + 0.5), xytext=(10, 0), textcoords='offset points', fontsize=7,
                            family=figure_font,
                            ha='left', va='center')
                ax.annotate(cost_label_str, xy=(0, i + 0.5), xytext=(-10, 0), textcoords='offset points', fontsize=7,
                            family=figure_font,
                            ha='right', va='center')
                ax.annotate('', xy=(0, i + 1), xytext=(0, i),
                            arrowprops=dict(linestyle=THEMES[theme]['failstyle'] if s.burns == [] else '-',
                                            **out_arrow), zorder=-10)
            xy += [(0, i)]

        # REV - Alterlate method that simply plots a line; not general
        # ax.plot(*zip(*xy), color=path_color, linewidth=path_width, zorder=-10, label=None)

        return fig